#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'so1n'
__date__ = '2020-02'
import asyncio
import logging
import time
from contextlib import contextmanager
from random import random
from typing import Iterator, NoReturn, Optional, Union

from aio_statsd.connection import Connection
from aio_statsd.pool import Pool
from aio_statsd.protocol import StatsdProtocol
from aio_statsd.protocol import DogStatsdProtocol
from aio_statsd.transport_layer_protocol import ProtocolFlag


class Client:

    def __init__(
            self,
            host: str = "localhost",
            port: int = 8125,
            protocol: ProtocolFlag = ProtocolFlag.udp,

            timeout: int = 0,
            debug: bool = False,
            close_timeout: int = 5,
            create_timeout: int = 5,
            read_timeout: float = 0.5,
            loop: Optional['asyncio.get_event_loop'] = None,
    ) -> NoReturn:
        self._queue: asyncio.Queue = asyncio.Queue()
        self._listen_future: Optional[asyncio.Future] = None
        self._join_future: Optional[asyncio.Future] = None

        self._is_listen: bool = False
        self._read_timeout: float = read_timeout
        self._close_timeout: float = close_timeout
        self._loop = loop if loop else asyncio.get_event_loop()

        self._conn_config_dict = {
            'host': host,
            'port': port,
            'protocol_flag': protocol,
            'debug': debug,
            'timeout': timeout,
            'create_timeout': create_timeout,
            'loop': self._loop
        }

        self.connection: Union[Pool, Connection, None] = None

    async def __aenter__(self) -> "Client":
        await self.connect()
        return self

    async def __aexit__(self, *args) -> NoReturn:
        await self.close()

    @property
    def is_closed(self):
        return not (self.connection and not self.connection.is_closed)

    async def connect(self) -> NoReturn:
        if not self.is_closed:
            raise ConnectionError(f'aiostatsd client already connected')
        self.connection = Connection(**self._conn_config_dict)
        await self._connect()

    async def create_pool(self, min_size: int = 1, max_size: int = 10):
        if not self.is_closed:
            raise ConnectionError(f'aiostatsd client already connected')
        self.connection = Pool(
            **self._conn_config_dict,
            min_size=min_size,
            max_size=max_size
        )
        await self._connect()

    async def _connect(self):
        await self.connection.connect()
        self._queue = asyncio.Queue()
        self._is_listen = True
        self._listen_future = asyncio.ensure_future(self._listen())
        logging.info(f'create aiostatsd client{self}')

    async def close(self) -> NoReturn:
        self._is_listen = False
        await self._listen_future
        self._listen_future = None

    async def _close(self):
        if not self.is_closed:
            try:
                await asyncio.wait_for(self._before_close(), timeout=self._close_timeout)
            except asyncio.TimeoutError:
                pass
            await self.connection.await_close()

    async def _before_close(self):
        while not self._queue.empty():
            await self._real_send()

    async def _listen(self) -> NoReturn:
        try:
            while self._is_listen:
                if not self._queue.empty():
                    await self._real_send()
                else:
                    await asyncio.sleep(0.01)
        except Exception as e:
            logging.error(f'aiostatsd listen status:{self._is_listen} error: {e}')
        finally:
            await self._close()

    async def _real_send(self) -> NoReturn:
        try:
            msg = await asyncio.wait_for(self._queue.get(), timeout=self._read_timeout)
        except asyncio.TimeoutError:
            logging.error(f'aiostatsd get by queue timeout')
        else:
            try:
                self.connection.sendto(msg)
            except Exception as e:
                logging.error(
                    f'connection:{self.connection}'
                    f' send msd error:{e}, drop msg:{msg}'
                )

    def send(self, msg: str) -> NoReturn:
        try:
            self._queue.put_nowait(msg)
        except Exception as e:
            logging.error(f'aiostatsd put:{msg} to queue error:{e}')


class GraphiteClient(Client):

    def __init__(
            self,
            host: str = "localhost",
            port: int = 2003,
            protocol: ProtocolFlag = ProtocolFlag.udp,

            timeout: int = 0,
            debug: bool = False,
            close_timeout: int = 5,
            create_timeout: int = 5,
            read_timeout: float = 0.5,
            loop: Optional['asyncio.get_event_loop'] = None
    ) -> NoReturn:
        super().__init__(
            host=host,
            port=port,
            protocol=protocol,
            timeout=timeout,
            debug=debug,
            close_timeout=close_timeout,
            create_timeout=create_timeout,
            read_timeout=read_timeout,
            loop=loop
        )

    def send_graphite(
            self,
            key: str,
            value: int = 0,
            timestamp: int = time.time(),
            interval: int = 10
    ) -> NoReturn:
        """interval: Multiple clients timestamp interval synchronization"""
        timestamp = int(timestamp) // interval * interval
        msg = "{} {} {}".format(key, value, timestamp)
        self.send(msg)


class StatsdClient(Client):

    def __init__(
            self,
            host: str = "localhost",
            port: int = 8125,
            protocol: ProtocolFlag = ProtocolFlag.udp,
            sample_rate: Union[float, int] = 1,

            timeout: int = 0,
            debug: bool = False,
            close_timeout: int = 5,
            create_timeout: int = 5,
            read_timeout: float = 0.5,
            loop: Optional['asyncio.get_event_loop'] = None
    ) -> NoReturn:
        super().__init__(
            host=host,
            port=port,
            protocol=protocol,
            timeout=timeout,
            debug=debug,
            close_timeout=close_timeout,
            create_timeout=create_timeout,
            read_timeout=read_timeout,
            loop=loop
        )
        self._sample_rate = sample_rate

    def send_statsd(
            self,
            statsd_protocol: 'StatsdProtocol',
            sample_rate: Union[int, float, None] = None
    ) -> NoReturn:
        msg = statsd_protocol.msg
        if '\n' in msg and sample_rate:
            logging.warning('Multi-Metric not support sample rate')
        else:
            sample_rate = sample_rate or self._sample_rate
            if sample_rate != 1 and random() > sample_rate:
                msg += f'|@{sample_rate}'
            else:
                return
        self.send(msg)

    def counter(
            self,
            key: str,
            value: int,
            sample_rate: Union[int, float, None] = None
    ) -> NoReturn:
        statsd_protocol = StatsdProtocol().counter(key, value)
        return self.send_statsd(statsd_protocol, sample_rate)

    def timer(
            self,
            key: str,
            value: Union[int, float],
            sample_rate: Union[int, float, None] = None
    ) -> NoReturn:
        statsd_protocol = StatsdProtocol().timer(key, value)
        return self.send_statsd(statsd_protocol, sample_rate)

    def gauge(
            self,
            key: str,
            value: int,
            sample_rate: Union[int, float, None] = None
    ) -> NoReturn:
        statsd_protocol = StatsdProtocol().gauge(key, value)
        return self.send_statsd(statsd_protocol, sample_rate)

    def sets(
            self,
            key: str,
            value: int,
            sample_rate: Union[int, float, None] = None
    ) -> NoReturn:
        statsd_protocol = StatsdProtocol().sets(key, value)
        return self.send_statsd(statsd_protocol, sample_rate)

    @contextmanager
    def timeit(
            self,
            key: str,
            sample_rate: Union[int, float, None] = None
    ) -> Iterator[None]:
        """
        Context manager for easily timing methods.
        """
        _loop = asyncio.get_event_loop()
        started_at = _loop.time()
        yield
        value = (_loop.time() - started_at) * 1000
        self.timer(key, value, sample_rate)


class DogStatsdClient(Client):

    def __init__(
            self,
            host: str = "localhost",
            port: int = 8125,
            protocol: ProtocolFlag = ProtocolFlag.udp,
            sample_rate: Union[float, int] = 1,

            timeout: int = 0,
            debug: bool = False,
            close_timeout: int = 5,
            create_timeout: int = 5,
            read_timeout: float = 0.5,
            loop: Optional['asyncio.get_event_loop'] = None
    ) -> NoReturn:
        super().__init__(
            host=host,
            port=port,
            protocol=protocol,
            timeout=timeout,
            debug=debug,
            close_timeout=close_timeout,
            create_timeout=create_timeout,
            read_timeout=read_timeout,
            loop=loop
        )
        self._sample_rate = sample_rate

    def send_dog_statsd(
            self,
            dog_statsd_protocol: 'DogStatsdProtocol',
            sample_rate: Union[int, float, None] = None
    ) -> NoReturn:
        for msg in dog_statsd_protocol.get_msg_list():
            sample_rate = sample_rate or self._sample_rate
            if sample_rate != 1 and random() > sample_rate:
                msg += f'|@{sample_rate}'
            else:
                continue
            self.send(msg)

    def gauge(
            self,
            key: str,
            value: int,
            sample_rate: Union[int, float, None] = None,
            tag_dict: Optional[dict] = None
    ) -> NoReturn:
        protocol = DogStatsdProtocol().gauge(key, value, tag_dict)
        return self.send_dog_statsd(protocol, sample_rate)

    def increment(
            self,
            key: str,
            value: int,
            sample_rate: Union[int, float, None] = None,
            tag_dict: Optional[dict] = None
    ) -> NoReturn:
        protocol = DogStatsdProtocol().increment(key, value, tag_dict)
        return self.send_dog_statsd(protocol, sample_rate)

    def decrement(
            self,
            key: str,
            value: int,
            sample_rate: Union[int, float, None] = None,
            tag_dict: Optional[dict] = None
    ) -> NoReturn:
        protocol = DogStatsdProtocol().increment(key, value, tag_dict)
        return self.send_dog_statsd(protocol, sample_rate)

    def timer(
            self,
            key: str,
            value: Union[int, float],
            sample_rate: Union[int, float, None] = None,
            tag_dict: Optional[dict] = None
    ) -> NoReturn:
        protocol = DogStatsdProtocol().timer(key, value, tag_dict)
        return self.send_dog_statsd(protocol, sample_rate)

    @contextmanager
    def timeit(
            self,
            key: str,
            sample_rate: Union[int, float, None] = None
    ) -> Iterator[None]:
        """
        Context manager for easily timing methods.
        """
        _loop = asyncio.get_event_loop()
        started_at = _loop.time()
        yield
        value = (_loop.time() - started_at) * 1000
        self.timer(key, value, sample_rate)

    def histogram(
            self,
            key: str,
            value: Union[int, float],
            sample_rate: Union[int, float, None] = None,
            tag_dict: Optional[dict] = None
    ) -> NoReturn:
        protocol = DogStatsdProtocol().histogram(key, value, tag_dict)
        return self.send_dog_statsd(protocol, sample_rate)

    def distribution(
            self,
            key: str,
            value: Union[int, float],
            sample_rate: Union[int, float, None] = None,
            tag_dict: Optional[dict] = None
    ) -> NoReturn:
        protocol = DogStatsdProtocol().distribution(key, value, tag_dict)
        return self.send_dog_statsd(protocol, sample_rate)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'so1n'
__date__ = '2020-02'
import logging
from typing import List, Optional


class StatsdProtocol(object):
    """learn more statsd protocol
    https://github.com/statsd/statsd/blob/master/docs/metric_types.md
    """
    def __init__(self, prefix: Optional[str] = None, mtu_limit: int = 1432):
        if prefix:
            self.msg = f'{prefix}.'
        else:
            self.msg = ''
        self._mtu_limit = mtu_limit

    def _msg_handle(self, msg: str):
        if self.msg == '':
            self.msg = msg
        self.msg += '\n' + msg
        if len(self.msg) > self._mtu_limit:
            logging.warning(
                f'msg length > mtu max length limit:{self._mtu_limit}'
                f'learn more:https://github.com/statsd/statsd/blob/master/docs/metric_types.md'
            )

    def counter(self, key: str, value: int) -> 'StatsdProtocol':
        self._msg_handle(f'{key}:{value}|c')
        return self

    def timer(self, key: str, value: int) -> 'StatsdProtocol':
        self._msg_handle(f'{key}:{value}|ms')
        return self

    def gauge(self, key: str, value: int) -> 'StatsdProtocol':
        self._msg_handle(f'{key}:{value}|g')
        return self

    def sets(self, key: str, value: int) -> 'StatsdProtocol':
        self._msg_handle(f'{key}:{value}|s')
        return self


class DogStatsdProtocol(object):

    def __init_t_(self, prefix: Optional[str] = None):
        if prefix:
            self._msg = f'{prefix}.'
        else:
            self._msg = ''
        self._cache: List[str] = []

    def get_msg_list(self) -> List[str]:
        return self._cache

    @staticmethod
    def build_tag(tag_dict: Optional[dict] = None):
        if not tag_dict:
            return ""
        return "|#" + ",".join(f"{k}:{v}" for k, v in tag_dict.items())

    def gauge(self, key: str, value: int, tag_dict: Optional[dict] = None) -> 'DogStatsdProtocol':
        msg = self._msg + f'{key}:{value}|g' + self.build_tag(tag_dict)
        self._cache.append(msg)
        return self

    def increment(self, key: str, value: int, tag_dict: Optional[dict] = None) -> 'DogStatsdProtocol':
        msg = self._msg + f'{key}:{value}|c' + self.build_tag(tag_dict)
        self._cache.append(msg)
        return self

    def decrement(self, key: str, value: int, tag_dict: Optional[dict] = None) -> 'DogStatsdProtocol':
        msg = self._msg + f'{key}:-{value}|c' + self.build_tag(tag_dict)
        self._cache.append(msg)
        return self

    def timer(self, key: str, value: int, tag_dict: Optional[dict] = None) -> 'DogStatsdProtocol':
        msg = self._msg + f'{key}:{value}|ms' + self.build_tag(tag_dict)
        self._cache.append(msg)
        return self

    def histogram(self, key: str, value: int, tag_dict: Optional[dict] = None) -> 'DogStatsdProtocol':
        msg = self._msg + f'{key}:{value}|h' + self.build_tag(tag_dict)
        self._cache.append(msg)
        return self

    def distribution(self, key: str, value: int, tag_dict: Optional[dict] = None) -> 'DogStatsdProtocol':
        msg = self._msg + f'{key}:{value}|d' + self.build_tag(tag_dict)
        self._cache.append(msg)
        return self



from aio_statsd.client import DogStatsdClient
from aio_statsd.client import GraphiteClient
from aio_statsd.client import StatsdClient
from aio_statsd.protocol import DogStatsdProtocol
from aio_statsd.protocol import StatsdProtocol

__all__ = [
    'DogStatsdClient', 'GraphiteClient', 'StatsdClient',
    'DogStatsdProtocol', 'StatsdProtocol',
    'ProtocolFlag'
]


name = "aio_statsd"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'so1n'
__date__ = '2020-02'
import asyncio
import logging

from typing import NoReturn
from aio_statsd.transport_layer_protocol import DatagramProtocol, TcpProtocol, ProtocolFlag


class Connection(object):
    def __init__(
            self,
            host: str,
            port: int,
            protocol_flag: ProtocolFlag,
            debug: bool,
            timeout: int,
            create_timeout: int,
            loop: 'asyncio.get_event_loop'
    ):
        self._debug: bool = debug
        self._protocol_flag: ProtocolFlag = protocol_flag
        self._create_timeout: int = create_timeout
        self._loop = loop

        self._connection_info = f'{protocol_flag}://{host}:{port}'
        if protocol_flag == ProtocolFlag.udp:
            self._connection: DatagramProtocol = DatagramProtocol(timeout=timeout)
            self._connection_proxy = self._loop.create_datagram_endpoint(
                lambda: self._connection,
                remote_addr=(host, port)
            )
        elif protocol_flag == ProtocolFlag.tcp:
            self._connection: TcpProtocol = TcpProtocol(timeout=timeout)
            self._connection_proxy = self._loop.create_connection(
                lambda: self._connection,
                host=host, port=port
            )
        else:
            raise ConnectionError(f'Not support protocol:{protocol_flag}')

        if debug:
            self.sendto = self._sendto_debug
        else:
            self.sendto = self._sendto

    async def connect(self) -> NoReturn:
        try:
            await asyncio.wait_for(self._connection_proxy, timeout=self._create_timeout)
        except asyncio.TimeoutError as e:
            raise TimeoutError(f'create connection:{self._connection_info} timeout') from e
        logging.debug(f'create connection:{self._connection_info}')

    def _sendto(self, data: str) -> NoReturn:
        if not self.is_closed:
            self._connection.send(bytes(data, encoding="utf8"))

    def _sendto_debug(self, data: str) -> NoReturn:
        if not self.is_closed:
            logging.debug(f'send msg:{data}')
            self._connection.send(bytes(data, encoding="utf8"))

    @property
    def is_closed(self) -> bool:
        return self._connection.is_closed

    def close(self):
        self._connection.close()

    async def wait_closed(self):
        await self._connection.wait_closed()

    async def await_close(self):
        await self._connection.await_close()
        logging.debug(f'close connection: {self._connection_info}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'so1n'
__date__ = '2020-07'
import asyncio
import logging
from enum import Enum

from typing import Optional, Union

__all__ = ['ProtocolFlag', 'TcpProtocol', 'DatagramProtocol']
logger = logging.getLogger()


class ProtocolFlag(Enum):
    tcp = 1
    udp = 2

    def __str__(self):
        return "%s" % self._name_

    def __repr__(self):
        return "%s" % self._name_


class _ProtocolMixin(asyncio.BaseProtocol):

    def __init__(self, timeout: int = 0) -> None:
        self._transport: Union[asyncio.Transport, asyncio.DatagramTransport, None] = None
        self._future: Optional[asyncio.Future] = None

        self._before_event = []
        self._after_event = []

        self._keep_alive_future: Optional[asyncio.Future] = None
        self._timeout = timeout
        self._loop = asyncio.get_event_loop()

        if self._timeout:
            self._before_event.append(self.create_keep_alive)
            self._after_event.append(self.cancel_keep_alive)

    def timeout_handle(self):
        raise asyncio.TimeoutError(f'No response data received within {self._timeout} seconds')

    def cancel_keep_alive(self):
        if self._keep_alive_future:
            self._keep_alive_future.cancel()
            self._keep_alive_future = None

    def create_keep_alive(self):
        self._keep_alive_future = self._loop.call_later(self._timeout, self.timeout_handle)

    async def await_close(self) -> None:
        self.close()
        await self.wait_closed()

    async def wait_closed(self):
        await self._future

    def close(self):
        if self._transport is None:
            return

        if not self._transport.is_closing():
            self._transport.close()
            self._transport = None

    @property
    def is_closed(self) -> bool:
        if not self._future:
            return True
        return self._future.done()

    def connection_made(self, transport):
        self._transport = transport
        self._future = asyncio.Future()

    def connection_lost(self, _exc):
        if self._future.done():
            return
        self._future.set_result(True)
        if _exc:
            raise ConnectionError from _exc

    def before_transport(self):
        if self._transport is None:
            raise ConnectionError('connection is close')
        for fn in self._before_event:
            fn()

    def after_transport(self, data, *args, **kwargs):
        logger.debug(f'receive data:{data} args:{args}, kwargs:{kwargs}')
        for fn in self._after_event:
            fn()


class TcpProtocol(asyncio.Protocol, _ProtocolMixin):

    def __init__(self, timeout: int = 0, loop=None):
        super().__init__(timeout)
        self._drain_waiter: Optional[asyncio.Future] = None
        self._paused: bool = False
        self._connection_lost: bool = False
        self._loop = loop if loop else asyncio.get_event_loop()

    def data_received(self, data):
        self.after_transport(data)

    def pause_reading(self):
        self._transport and self._transport.pause_reading()
        assert not self._paused
        self._paused = True
        if self._loop.get_debug():
            logger.debug("%r pauses writing", self)

    def resume_reading(self):
        self._transport and self._transport.resume_reading()
        self._paused = False
        if self._loop.get_debug():
            logger.debug("%r resumes writing", self)

        if self._drain_waiter is not None:
            if not self._drain_waiter.done():
                self._drain_waiter.set_result(None)
            self._drain_waiter = None

    def connection_lost(self, exc):
        super().connection_lost(exc)
        self._connection_lost = True
        # Wake up the writer if currently paused.
        if not self._paused:
            return

        if self._drain_waiter is None:
            return
        if self._drain_waiter.done():
            return
        if exc is None:
            self._drain_waiter.set_result(None)
        else:
            self._drain_waiter.set_exception(exc)
        self._drain_waiter = None

    async def drain(self):
        if self._connection_lost:
            raise ConnectionResetError('Connection lost')
        if not self._paused:
            return

        assert self._drain_waiter is None or self._drain_waiter.cancelled()
        self._drain_waiter = asyncio.Future()
        await self._drain_waiter

    def eof_received(self):
        logger.debug(f'{self} receive `EOF` flag')
        self.close()

    def send(self, data: bytes) -> None:
        self.before_transport()
        self._transport.write(data)


class DatagramProtocol(asyncio.DatagramProtocol, _ProtocolMixin):

    def datagram_received(self, data: bytes, peer_name: str, *arg):
        self.after_transport(data, peer_name)

    def error_received(self, exc):
        self.close()
        raise exc

    def send(self, data: bytes) -> None:
        self.before_transport()
        self._transport.sendto(data)



"""

"""
import asyncio
import logging

from collections import deque
from typing import Coroutine, Deque, List, Set, Optional

from aio_statsd.connection import Connection
from aio_statsd.transport_layer_protocol import ProtocolFlag


class CloseEvent:
    def __init__(self, on_close):
        self._close_init = asyncio.Event()
        self._close_done = asyncio.Event()
        self._on_close = on_close

    async def wait(self):
        await self._close_init.wait()
        await self._close_done.wait()

    def is_set(self):
        return self._close_done.is_set() or self._close_init.is_set()

    def set(self):
        if self._close_init.is_set():
            return

        task = asyncio.ensure_future(self._on_close())
        task.add_done_callback(self._cleanup)
        self._close_init.set()

    def _cleanup(self, *args):
        self._on_close = None
        self._close_done.set()


class Pool(object):
    """Simple connection pool, design like aioredis.pool"""
    def __init__(
            self,
            host: str,
            port: int,
            protocol_flag: ProtocolFlag,
            debug: bool,
            timeout: int,
            create_timeout: int,
            loop: 'asyncio.get_event_loop',
            min_size: int = 1,
            max_size: int = 10,
    ):
        self._host: str = host
        self._port: int = port
        self._protocol: ProtocolFlag = protocol_flag
        self._debug: bool = debug
        self._timeout: int = timeout
        self._create_timeout: int = create_timeout
        self._loop = loop if loop else asyncio.get_event_loop()

        self._min_size = min_size
        self._max_size = max_size

        self._is_closed: bool = True
        self._cond = asyncio.Condition()
        self._close_state = CloseEvent(self._do_close)
        self._pool: Deque[Connection] = deque(maxlen=max_size)
        self._used_set: Set[Connection] = set()

    @property
    def free_size(self):
        """Current number of free connections."""
        return len(self._pool)

    @property
    def size(self):
        return self.free_size + self.use_size

    @property
    def use_size(self):
        return len(self._used_set)

    def _drop_closed(self):
        for i in range(self.free_size):
            conn = self._pool[0]
            if conn.is_closed:
                self._pool.popleft()
            else:
                self._pool.rotate(-1)

    def sendto(self, msg: str):
        asyncio.ensure_future(self.send_await(msg))

    async def send_await(self, msg: str):
        conn = await self.acquire()
        try:
            return conn.sendto(msg)
        finally:
            self.release(conn)

    async def _do_close(self):
        async with self._cond:
            waiter_list: List[Coroutine] = []
            while self._pool:
                waiter_list.append(self._pool.popleft().await_close())
            for conn in self._used_set:
                waiter_list.append(conn.await_close())
            await asyncio.gather(*waiter_list)
            logging.debug("Closed %d connection(s)", len(waiter_list))

    @property
    def is_closed(self):
        return self._close_state.is_set()

    async def wait_closed(self):
        await self._close_state.wait()

    async def await_close(self):
        self.close()
        await self.wait_closed()

    def close(self):
        if not self._close_state.is_set():
            self._close_state.set()

    async def connect(self):
        self._drop_closed()
        while self.size < self._max_size:
            try:
                conn = Connection(
                    self._host,
                    self._port,
                    self._protocol,
                    self._debug,
                    self._timeout,
                    self._create_timeout,
                    self._loop
                )
                await conn.connect()
                self._pool.append(conn)
            finally:
                self._drop_closed()
                await asyncio.sleep(0.01)

    async def acquire(self):
        async with self._cond:
            while True:
                await self.connect()
                if self.free_size:
                    conn = self._pool.popleft()
                    self._used_set.add(conn)
                    return conn
                else:
                    await self._cond.wait()

    def release(self, conn: Connection):
        if conn not in self._used_set:
            raise RuntimeError("connection not in user pool")
        self._used_set.remove(conn)
        if conn.is_closed:
            asyncio.ensure_future(self._wakeup(conn))
        elif self.free_size < self._max_size:
            self._pool.append(conn)
            asyncio.ensure_future(self._wakeup())
        else:
            asyncio.ensure_future(self._wakeup(conn))

    async def _wakeup(self, conn: Optional[Connection] = None):
        async with self._cond:
            self._cond.notify()
        if conn is not None:
            await conn.close()
